@@ -52,7 +52,6 @@ import java.io.IOException;
 import java.net.URI;
 import java.net.URISyntaxException;
 import java.util.Collections;
-import java.util.Date;
 import java.util.Map;
 import java.util.concurrent.TimeUnit;
 import javax.servlet.http.Cookie;
@@ -215,7 +214,7 @@ public class OIDCWebSSOProviderImpl implements OIDCWebSSOProvider {
 
     private String getRedirectLocationForSession(HttpServletRequest request, Session session) {
         LOG.trace("getRedirectLocationForSession(HttpServletRequest request: {}, Session session: {})", request.getRequestURI(), session.getSessionID());
-        return OIDCTools.buildFrontendRedirectLocation(session, OIDCTools.getUIWebPath(this.loginConfiguration, this.backend.getBackendConfig()), request.getParameter(OIDCTools.PARAM_DEEP_LINK));
+        return OIDCTools.buildFrontendRedirectLocation(session, OIDCTools.getUIWebPath(this.loginConfiguration, this.backend.getBackendConfig()), OIDCTools.getDeepLink(request));
     }
 
     private String buildLoginRequest(State state, Nonce nonce, HttpServletRequest request) throws OXException {
@@ -238,7 +237,7 @@ public class OIDCWebSSOProviderImpl implements OIDCWebSSOProvider {
 
     private void addAuthRequestToStateManager(State state, Nonce nonce, HttpServletRequest request)  throws OXException {
         LOG.trace("addAuthRequestToStateManager(State state: {}, Nonce nonce: {}, HttpServletRequest request: {})", state.getValue(), nonce.getValue(), request.getRequestURI());
-        String deepLink = request.getParameter("hash");
+        String deepLink = OIDCTools.getDeepLink(request);
         String uiClientID = OIDCTools.getUiClient(request);
         String hostname = OIDCTools.getDomainName(request, services.getOptionalService(HostnameService.class));
         Map<String, String> additionalClientInformation = Collections.emptyMap();
@@ -256,9 +255,11 @@ public class OIDCWebSSOProviderImpl implements OIDCWebSSOProvider {
     @Override
     public void authenticateUser(HttpServletRequest request, HttpServletResponse response) throws OXException{
         LOG.trace("authenticateUser(HttpServletRequest request: {}, HttpServletResponse response)", request.getRequestURI());
+
+        String code = request.getParameter("code");
         AuthenticationRequestInfo storedRequestInformation = this.stateManagement.getAndRemoveAuthenticationInfo(request.getParameter("state"));
 
-        if (storedRequestInformation == null) {
+        if (storedRequestInformation == null && Strings.isEmpty(code)) {
             throw OIDCExceptionCode.INVALID_AUTHENTICATION_STATE_NO_USER.create();
         }
 
@@ -269,16 +270,22 @@ public class OIDCWebSSOProviderImpl implements OIDCWebSSOProvider {
             if (null == idToken) {
                 throw OXException.general("Missing IDToken");
             }
-            IDTokenClaimsSet validTokenResponse = this.backend.validateIdToken(idToken, storedRequestInformation.getNonce());
+
+            String nonce = null;
+            if (storedRequestInformation != null) {
+                nonce = storedRequestInformation.getNonce();
+            }
+            IDTokenClaimsSet validTokenResponse = this.backend.validateIdToken(idToken, nonce);
             if (validTokenResponse == null) {
                 throw OIDCExceptionCode.IDTOKEN_GATHERING_ERROR.create("IDToken validation failed, no claim set could be extracted");
             }
             this.sendLoginRequestToServer(request, response, tokenResponse, storedRequestInformation);
         } catch (OXException e) {
-            if (e.getExceptionCode() != OIDCExceptionCode.IDTOKEN_GATHERING_ERROR) {
-                throw OIDCExceptionCode.IDTOKEN_GATHERING_ERROR.create(e, e.getMessage());
+            if (e.getExceptionCode() instanceof OIDCExceptionCode) {
+                throw e;
             }
-            throw e;
+
+            throw OIDCExceptionCode.IDTOKEN_GATHERING_ERROR.create(e, e.getMessage());
         }
     }
 
@@ -328,18 +335,27 @@ public class OIDCWebSSOProviderImpl implements OIDCWebSSOProvider {
     private void sendLoginRequestToServer(HttpServletRequest request, HttpServletResponse response, OIDCTokenResponse tokenResponse, AuthenticationRequestInfo storedRequestInformation) throws OXException {
         LOG.trace("sendLoginRequestToServer(HttpServletRequest request: {}, HttpServletResponse response, OIDCTokenResponse tokenResponse: {}, AuthenticationRequestInfo storedRequestInformation: {})",
             request.getRequestURI(), tokenResponse.getOIDCTokens().toJSONObject().toJSONString(), storedRequestInformation);
+        BearerAccessToken bearerAccessToken = tokenResponse.getTokens().getBearerAccessToken();
+        if (bearerAccessToken == null) {
+            // could occur, if token response contains access token of a different type than "bearer"
+            throw OIDCExceptionCode.UNABLE_TO_PARSE_RESPONSE_FROM_IDP.create("get bearer access token");
+        }
+
         AuthenticationInfo authInfo = this.backend.resolveAuthenticationResponse(request, tokenResponse);
         authInfo.setProperty(OIDCTools.IDTOKEN, tokenResponse.getOIDCTokens().getIDTokenString());
-        BearerAccessToken bearerAccessToken = tokenResponse.getTokens().getBearerAccessToken();
-        RefreshToken refreshToken = tokenResponse.getTokens().getRefreshToken();
-        if (bearerAccessToken != null && refreshToken != null) {
-            authInfo.setProperty(OIDCTools.ACCESS_TOKEN, bearerAccessToken.getValue());
-            authInfo.setProperty(OIDCTools.REFRESH_TOKEN, refreshToken.getValue());
-            long expiryDate = new Date().getTime();
-            expiryDate += bearerAccessToken.getLifetime() * 1000;
-            authInfo.setProperty(OIDCTools.ACCESS_TOKEN_EXPIRY, String.valueOf(expiryDate));
+        authInfo.setProperty(OIDCTools.ACCESS_TOKEN, bearerAccessToken.getValue());
+        long expiresIn = bearerAccessToken.getLifetime();
+        if (expiresIn > 0) {
+            authInfo.setProperty(OIDCTools.ACCESS_TOKEN_EXPIRY, String.valueOf(OIDCTools.expiresInToDate(expiresIn).getTime()));
         }
 
+        RefreshToken refreshToken = tokenResponse.getTokens().getRefreshToken();
+        if (refreshToken != null) {
+            authInfo.setProperty(OIDCTools.REFRESH_TOKEN, refreshToken.getValue());
+        }
+
+        String domainName = OIDCTools.getDomainName(request, services.getOptionalService(HostnameService.class));
+
         String sessionToken = sessionReservationService.reserveSessionFor(
             authInfo.getUserId(),
             authInfo.getContextId(),
@@ -350,19 +366,19 @@ public class OIDCWebSSOProviderImpl implements OIDCWebSSOProvider {
         try {
             URIBuilder redirectLocation = new URIBuilder()
                 .setScheme(OIDCTools.getRedirectScheme(request))
-                .setHost(storedRequestInformation.getDomainName())
+                .setHost(storedRequestInformation != null ? storedRequestInformation.getDomainName() : domainName)
                 .setPath(OIDCTools.getRedirectPathPrefix() + "login")
                 .setParameter(OIDCTools.SESSION_TOKEN, sessionToken)
                 .setParameter(LoginServlet.PARAMETER_ACTION, OIDCTools.OIDC_LOGIN + OIDCTools.getPathString(this.backend.getPath()))
                 .setParameter(OIDCTools.PARAM_SHARD, SessionUtility.getShardCookieValue());
             Tools.disableCaching(response);
 
-            String deepLink = storedRequestInformation.getDeepLink();
+            String deepLink = storedRequestInformation != null ? storedRequestInformation.getDeepLink() : null;
             if (deepLink != null) {
                 redirectLocation.setParameter(OIDCTools.PARAM_DEEP_LINK, deepLink);
             }
 
-            String clientID = storedRequestInformation.getUiClientID();
+            String clientID = storedRequestInformation != null ? storedRequestInformation.getUiClientID() : null;
             if (clientID != null) {
                 redirectLocation.setParameter(LoginFields.CLIENT_PARAM, clientID);
             }
@@ -400,14 +416,13 @@ public class OIDCWebSSOProviderImpl implements OIDCWebSSOProvider {
 
     private String getResumeURL(HttpServletRequest request, HttpServletResponse response, String domainName, Session session) {
         String redirectURI = "";
-        String path = OIDCTools.buildFrontendRedirectLocation(session, OIDCTools.getUIWebPath(this.loginConfiguration, this.backend.getBackendConfig()), request.getParameter("hash"));
+        String location = OIDCTools.buildFrontendRedirectLocation(session, OIDCTools.getUIWebPath(this.loginConfiguration, this.backend.getBackendConfig()), OIDCTools.getDeepLink(request));
         URIBuilder redirectLocation = new URIBuilder()
             .setScheme(OIDCTools.getRedirectScheme(request))
-            .setHost(domainName)
-            .setPath(path);
+            .setHost(domainName);
         Tools.disableCaching(response);
         try {
-            redirectURI = redirectLocation.build().toString();
+            redirectURI = redirectLocation.build().toString() + location;
         } catch (URISyntaxException e) {
             //should not happen
             LOG.error("", e);
