@@ -56,33 +56,47 @@ import java.io.PrintWriter;
 import java.net.URI;
 import java.net.URISyntaxException;
 import java.util.Collection;
+import java.util.Date;
+import java.util.HashMap;
 import java.util.Map;
+import java.util.Optional;
+import java.util.concurrent.TimeUnit;
 import javax.servlet.http.Cookie;
 import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletResponse;
+import org.apache.http.client.utils.URIBuilder;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
-import com.openexchange.ajax.AJAXUtility;
+import com.nimbusds.oauth2.sdk.token.AccessToken;
+import com.nimbusds.oauth2.sdk.token.RefreshToken;
+import com.nimbusds.oauth2.sdk.token.Tokens;
 import com.openexchange.ajax.LoginServlet;
 import com.openexchange.ajax.SessionUtility;
 import com.openexchange.ajax.login.HashCalculator;
 import com.openexchange.ajax.login.LoginConfiguration;
 import com.openexchange.ajax.login.LoginTools;
+import com.openexchange.authentication.Authenticated;
+import com.openexchange.authentication.DefaultAuthenticated;
 import com.openexchange.dispatcher.DispatcherPrefixService;
 import com.openexchange.exception.OXException;
+import com.openexchange.groupware.contexts.Context;
 import com.openexchange.groupware.notify.hostname.HostnameService;
 import com.openexchange.java.Charsets;
 import com.openexchange.java.Strings;
+import com.openexchange.oidc.AuthenticationInfo;
 import com.openexchange.oidc.OIDCBackend;
 import com.openexchange.oidc.OIDCBackendConfig;
 import com.openexchange.oidc.OIDCExceptionCode;
 import com.openexchange.oidc.osgi.Services;
 import com.openexchange.session.Session;
+import com.openexchange.session.oauth.OAuthTokens;
+import com.openexchange.session.oauth.TokenRefreshConfig;
 import com.openexchange.sessiond.SessionExceptionCodes;
 import com.openexchange.sessiond.SessionFilter;
 import com.openexchange.sessiond.SessiondService;
 import com.openexchange.tools.servlet.http.Cookies;
 import com.openexchange.tools.servlet.http.Tools;
+import com.openexchange.user.User;
 
 /**
  * {@link OIDCTools}
@@ -126,10 +140,23 @@ public class OIDCTools {
 
     public static final String DEFAULT_BACKEND_PATH = "oidc";
 
-    public static final String PARAM_DEEP_LINK = "hash";
+    public static final String PARAM_DEEP_LINK = "uriFragment";
+
+    @Deprecated
+    public static final String PARAM_DEEP_LINK_ALT = "hash";
 
     public static final String PARAM_SHARD = "shard";
 
+    /**
+     * Map key to preserve user login info
+     */
+    public static final String USER_LOGIN_INFO = "user_login_info";
+
+    /**
+     * Map key to preserve context login info
+     */
+    public static final String CONTEXT_LOGIN_INFO = "context_login_info";
+
     public static String getPathString(String path) {
         if (Strings.isEmpty(path)) {
             return "";
@@ -145,16 +172,10 @@ public class OIDCTools {
      * @param deeplink The deeplink, can be <code>null</code>
      */
     public static String buildFrontendRedirectLocation(Session session, String uiWebPath, String deeplink) {
-        LOG.trace("buildFrontendRedirectLocation(Session session: {}, String uiWebPath: {})", session.getSessionID(), uiWebPath);
-        String retval = uiWebPath;
-
-        // Prevent HTTP response splitting.
-        retval = retval.replaceAll("[\n\r]", "");
-        retval = LoginTools.addFragmentParameter(retval, PARAMETER_SESSION, session.getSessionID());
-        if (!Strings.isEmpty(deeplink)) {
-            retval += "&" + AJAXUtility.encodeUrl(deeplink.substring(1), true);
-        }
-        return retval;
+        LOG.trace("buildFrontendRedirectLocation(Session session: {}, String uiWebPath: {}, String deeplink: {})", session.getSessionID(), uiWebPath, deeplink);
+        URIBuilder location = new URIBuilder();
+        setRedirectLocation(location, session, uiWebPath, deeplink);
+        return location.toString();
     }
 
     /**
@@ -371,6 +392,21 @@ public class OIDCTools {
         return prefixService.getPrefix();
     }
 
+    /**
+     * Gets a deep link (URI fragment hash to be appended to final UI redirect) from
+     * the client request.
+     *
+     * @param request The request
+     * @return The value to preserve or <code>null</code>
+     */
+    public static String getDeepLink(HttpServletRequest request) {
+        String value = request.getParameter(PARAM_DEEP_LINK);
+        if (value == null) {
+            value = request.getParameter(PARAM_DEEP_LINK_ALT);
+        }
+        return value;
+    }
+
     /**
      * Load a session from the given {@link Cookie}.
      *
@@ -401,4 +437,207 @@ public class OIDCTools {
 
         return session;
     }
+
+    /**
+     * Creates an {@link Authenticated} instance based on the given context, user and {@link AuthenticationInfo}
+     * properties.
+     */
+    public static Authenticated getDefaultAuthenticated(final Context context, final User user, Map<String, String> state) {
+        LOG.trace("getDefaultAuthenticated(final Context context: {}, final User user: {})", I(context.getContextId()), I(user.getId()));
+        String contextInfo = state.get(OIDCTools.CONTEXT_LOGIN_INFO);
+        if (Strings.isEmpty(contextInfo)) {
+            contextInfo = context.getLoginInfo()[0];
+        }
+
+        String userInfo = state.get(OIDCTools.USER_LOGIN_INFO);
+        if (Strings.isEmpty(userInfo)) {
+            userInfo = user.getLoginInfo();
+        }
+
+        return new DefaultAuthenticated(contextInfo, userInfo);
+    }
+
+    /**
+     * Adds OIDC specific parameters to a session instance. Any parameter must be contained in the provided
+     * map. The following parameters and keys are allowed/expected:
+     * <p>
+     * <ul>
+     *   <li>ID token: {@link OIDCTools#IDTOKEN}</li>
+     *   <li>access token: {@link OIDCTools#ACCESS_TOKEN}</li>
+     *   <li>refresh token: {@link OIDCTools#REFRESH_TOKEN}</li>
+     *   <li>access token expiry: {@link OIDCTools#ACCESS_TOKEN_EXPIRY}</li>
+     *   <li>backend path: {@link OIDCTools#BACKEND_PATH}</li>
+     *   <li>last token refresh timestamp: {@link OIDCTools#LAST_REFRESH_TIMESTAMP}</li>
+     * </ul>
+     * <p>
+     */
+    public static void setSessionParameters(Session session, Map<String, String> params) {
+        OIDCTools.addParameterToSession(session, params, OIDCTools.IDTOKEN, OIDCTools.IDTOKEN);
+        OIDCTools.addParameterToSession(session, params, OIDCTools.ACCESS_TOKEN, Session.PARAM_OAUTH_ACCESS_TOKEN);
+        OIDCTools.addParameterToSession(session, params, OIDCTools.REFRESH_TOKEN, Session.PARAM_OAUTH_REFRESH_TOKEN);
+        OIDCTools.addParameterToSession(session, params, OIDCTools.ACCESS_TOKEN_EXPIRY, Session.PARAM_OAUTH_ACCESS_TOKEN_EXPIRY_DATE);
+
+        // preserve empty string which typically indicates the default backend
+        String backendPath = params.get(OIDCTools.BACKEND_PATH);
+        if (backendPath != null) {
+            session.setParameter(OIDCTools.BACKEND_PATH, backendPath);
+        }
+    }
+
+    /**
+     * Converts a token map into an {@link OAuthTokens} instance. The following parameters and keys are allowed/expected:
+     * <p>
+     * <ul>
+     *   <li>access token: {@link OIDCTools#ACCESS_TOKEN}</li>
+     *   <li>refresh token: {@link OIDCTools#REFRESH_TOKEN}</li>
+     *   <li>access token expiry: {@link OIDCTools#ACCESS_TOKEN_EXPIRY} - milliseconds since January 1, 1970, 00:00:00 GMT</li>
+     * </ul>
+     *
+     * @param params
+     */
+    public static Optional<OAuthTokens> convertTokenMap(Map<String, String> params) {
+        String accessToken = params.get(OIDCTools.ACCESS_TOKEN);
+        String expiryString = params.get(OIDCTools.ACCESS_TOKEN_EXPIRY);
+        String refreshToken = params.get(OIDCTools.REFRESH_TOKEN);
+
+        if (accessToken == null) {
+            return Optional.empty();
+        }
+
+        Date expiryDate = null;
+        if (expiryString != null) {
+            try {
+                long expiryMillis = Long.parseLong(expiryString);
+                expiryDate = new Date(expiryMillis);
+            } catch (NumberFormatException e) {
+                LOG.warn("Invalid entry in OAuth parameter map: {} => \"{}\"", OIDCTools.ACCESS_TOKEN_EXPIRY, expiryString);
+            }
+        }
+
+        return Optional.of(new OAuthTokens(accessToken, expiryDate, refreshToken));
+    }
+
+
+    /**
+     * Gets OIDC specific parameters from a session instance. The following parameters and keys are extracted if available:
+     * <p>
+     * <ul>
+     *   <li>ID token: {@link OIDCTools#IDTOKEN}</li>
+     *   <li>access token: {@link OIDCTools#ACCESS_TOKEN}</li>
+     *   <li>refresh token: {@link OIDCTools#REFRESH_TOKEN}</li>
+     *   <li>access token expiry: {@link OIDCTools#ACCESS_TOKEN_EXPIRY}</li>
+     *   <li>backend path: {@link OIDCTools#BACKEND_PATH}</li>
+     *   <li>last token refresh timestamp: {@link OIDCTools#LAST_REFRESH_TIMESTAMP}</li>
+     * </ul>
+     * <p>
+     */
+    public static Map<String, String> getSessionParameters(Session session) {
+        Map<String, String> params = new HashMap<String, String>(6, 1.0f);
+        Object parameter = session.getParameter(OIDCTools.IDTOKEN);
+        if (parameter instanceof String) {
+            params.put(IDTOKEN, (String) parameter);
+        }
+        parameter = session.getParameter(Session.PARAM_OAUTH_ACCESS_TOKEN);
+        if (parameter instanceof String) {
+            params.put(ACCESS_TOKEN, (String) parameter);
+        }
+        parameter = session.getParameter(Session.PARAM_OAUTH_REFRESH_TOKEN);
+        if (parameter instanceof String) {
+            params.put(REFRESH_TOKEN, (String) parameter);
+        }
+        parameter = session.getParameter(Session.PARAM_OAUTH_ACCESS_TOKEN_EXPIRY_DATE);
+        if (parameter instanceof String) {
+            params.put(ACCESS_TOKEN_EXPIRY, (String) parameter);
+        }
+        parameter = session.getParameter(BACKEND_PATH);
+        if (parameter instanceof String) {
+            params.put(BACKEND_PATH, (String) parameter);
+        }
+        return params;
+    }
+
+    /**
+     * Converts a {@link Tokens} instance to an {@link OAuthTokens} instance.
+     *
+     * @param tokens
+     */
+    public static OAuthTokens convertNimbusTokens(Tokens tokens) {
+        AccessToken accessToken = tokens.getAccessToken();
+        long lifetime = accessToken.getLifetime();
+        Date expiryDate = null;
+        if (lifetime > 0) {
+            long expiryDateMillis = System.currentTimeMillis();
+            expiryDateMillis += lifetime * 1000;
+            expiryDate = new Date(expiryDateMillis);
+        }
+
+        RefreshToken refreshToken = tokens.getRefreshToken();
+        String refreshTokenValue = null;
+        if (refreshToken != null) {
+            refreshTokenValue = refreshToken.getValue();
+        }
+
+        return new OAuthTokens(tokens.getAccessToken().getValue(), expiryDate, refreshTokenValue);
+    }
+
+    /**
+     * Gets a {@link TokenRefreshConfig} instance based on backend config
+     *
+     * @param config
+     */
+    public static TokenRefreshConfig getTokenRefreshConfig(OIDCBackendConfig config) {
+        return TokenRefreshConfig.newBuilder()
+            .setLockTimeout(config.getTokenLockTimeoutSeconds(), TimeUnit.SECONDS)
+            .setRefreshThreshold(config.getOauthRefreshTime(), TimeUnit.MILLISECONDS)
+            .setTryRecoverStoredTokens(config.tryRecoverStoredTokens())
+            .build();
+    }
+
+    /**
+     * Converts an OAuth access tokens <code>expires_in</code> property to a {@link Date}.
+     *
+     * @param expiresIn The <code>expires_in</code> value in seconds
+     * @return The date
+     */
+    public static Date expiresInToDate(long expiresIn) {
+        return new Date(System.currentTimeMillis() + (expiresIn * 1000));
+    }
+
+    /**
+     * Sets path, query and fragment at the passed {@link URIBuilder} instance to create
+     * an URI that can be used to bootstrap a frontend session with optional deep link.
+     *
+     * Path and deep link are sanitized before being appended. They are expected to be raw
+     * (non-URI-encoded) Strings.
+     *
+     * @param uri The URI builder
+     * @param session The session to jump into
+     * @param path The URI path
+     * @param deepLink The deep link argument which is appended to the URI fragment as <code>&[deepLink]</code>
+     */
+    private static void setRedirectLocation(URIBuilder uri, Session session, String path, String deepLink) {
+        uri.setPath(path.replaceAll("[\n\r]", ""));
+
+        StringBuilder fragment = new StringBuilder(PARAMETER_SESSION).append('=').append(session.getSessionID());
+        fragment.append(sanitizeDeepLinkFragment(deepLink));
+        uri.setFragment(fragment.toString());
+    }
+
+    private static final String sanitizeDeepLinkFragment(String uriFragment) {
+        if (uriFragment == null) {
+            return "";
+        }
+
+        uriFragment = uriFragment.replaceAll("[\n\r]", "");
+        while (uriFragment.length() > 0 && (uriFragment.charAt(0) == '#' || uriFragment.charAt(0) == '&')) {
+            uriFragment = uriFragment.substring(1);
+        }
+
+        if (uriFragment.length() > 0) {
+            uriFragment = "&" + uriFragment;
+        }
+
+        return uriFragment;
+    }
+
 }
