@@ -53,6 +53,7 @@ import java.util.List;
 import java.util.Map;
 import com.google.common.collect.ImmutableMap;
 import com.openexchange.authentication.AuthenticationService;
+import com.openexchange.authentication.NamePart;
 
 /**
  * {@link OIDCBackendConfig}
@@ -124,14 +125,6 @@ public interface OIDCBackendConfig {
      */
     String getClientID();
 
-    /**
-     * Get the redirect URI that should be used by the frontend to start the authentication
-     * process. Pointing to the init-Servlet. {@link InitService}
-     *
-     * @return the redirect URI
-     */
-    String getRpRedirectURIInit();
-
     /**
      * Get the redirect URI that the OP should redirect to after token generation. Pointing to
      * the authentication servlet. {@link AuthenticationService}
@@ -176,7 +169,8 @@ public interface OIDCBackendConfig {
     String getJWSAlgortihm();
 
     /**
-     * The used OIDC scope.
+     * The scope to request during OIDC authorization flow. This is a space-separated list
+     * of scope values, e.g. <code>openid offline</<code>. Scope values are case-sensitive!
      *
      * @return the scope
      */
@@ -240,10 +234,30 @@ public interface OIDCBackendConfig {
     boolean isAutologinEnabled();
 
     /**
-     * How long before an Oauth Access token expires a new token is supposed to be
-     * requested. Time in milliseconds.
+     * How long before an OAuth access token expires a new token is supposed to be
+     * requested, usually by exchanging a refresh token. In case refresh tokens are
+     * not issued, this value should be 0 to align max. allowed session use time to
+     * access token expiry.
+     * <p>
+     * <strong>Note that OAuth tokens contained in token responses are always set
+     * as session parameters and internally validated on each request. If the
+     * Authorization Server issues at least an access token with expiry date, this
+     * date will determine how long the session can be used!</strong>
+     * <p>A session containing an access token will be terminated on the first request
+     * that happens after <code>expires_in - (getOauthRefreshTime() / 1000)</code>
+     * seconds.
+     * <p>
+     * If a refresh token is contained in the token response, access tokens will be
+     * refreshed during the first request that happens after <code>(expires_in - (getOauthRefreshTime() / 1000))</code>
+     * seconds. A failed refreshed due an invalid/revoked refresh token or an
+     * <code>invalid_grant</code> response will lead to the session being terminated.
+     * <p>
+     * If <code>expires_in</code> is not contained in the token response, the session
+     * will live as long as the configured short-term session lifetime, no matter
+     * if a refresh token is contained or not. The access token will never be refreshed
+     * in that case.
      *
-     * @return the time
+     * @return Time in milliseconds.
      */
     int getOauthRefreshTime();
 
@@ -267,7 +281,7 @@ public interface OIDCBackendConfig {
      * @return
      */
     List<String> getHosts();
-    
+
     /**
      * Load the redirect URI that should be used if a server error occurs that can be ignored.
      * For example a redirect to the IDP for another login attempt.
@@ -275,4 +289,60 @@ public interface OIDCBackendConfig {
      * @return
      */
     String getFailureRedirect();
+
+    /**
+     * Gets the name of the ID token claim that will be used by the
+     * default backend to resolve a context.
+     */
+    String getContextLookupClaim();
+
+    /**
+     * Gets the {@link NamePart} of the ID token claim value used for
+     * determining the context of a user.
+     *
+     * @see #getContextLookupClaim()
+     */
+    NamePart getContextLookupNamePart();
+
+    /**
+     * Gets the name of the ID token claim that will be used by the
+     * default backend to resolve a user.
+     *
+     */
+    String getUserLookupClaim();
+
+    /**
+     * Gets the {@link NamePart} of the ID token claim value used for
+     * determining a user.
+     *
+     * @see #getUserLookupClaim()
+     */
+    NamePart getUserLookupNamePart();
+
+    /**
+     * Gets the {@link NamePart} to be used for an issued Resource
+     * Owner Password Credentials Grant ({@link https://tools.ietf.org/html/rfc6749#section-4.3}).
+     * The part is taken from the user-provided login name.
+     *
+     * @see OIDCProperty#enablePasswordGrant
+     */
+    NamePart getPasswordGrantUserNamePart();
+
+    /**
+     * Lock timeout before giving up trying to refresh an access token for
+     * a session. If multiple threads try to check or refresh the access token
+     * at the same time, only one gets a lock and blocks the others. In case
+     * of a timeout, this is logged as a temporary issue and the request continued
+     * as usual.
+     */
+    long getTokenLockTimeoutSeconds();
+
+    /**
+     * Gets whether token refresh should try to recover valid tokens from
+     * the session instance that is present in {@link SessionStorageService}.
+     * This is only tried as a fall-back, after token refresh failed with an
+     * {@code invalid_grant} error.
+     */
+    boolean tryRecoverStoredTokens();
+
 }
